@@ -3,14 +3,17 @@ try:
 except ImportError:
     OrderedDict = dict
 from collections import namedtuple
+from inspect import isclass
 import re
 
 from peewee import *
+from peewee import _StringField
+from peewee import text_type
 try:
-    from MySQLdb.constants import FIELD_TYPE
+    from pymysql.constants import FIELD_TYPE
 except ImportError:
     try:
-        from pymysql.constants import FIELD_TYPE
+        from MySQLdb.constants import FIELD_TYPE
     except ImportError:
         FIELD_TYPE = None
 try:
@@ -34,18 +37,21 @@ class Column(object):
     """
     Store metadata about a database column.
     """
-    primary_key_types = (IntegerField, PrimaryKeyField)
+    primary_key_types = (IntegerField, AutoField)
 
     def __init__(self, name, field_class, raw_column_type, nullable,
-                 primary_key=False, db_column=None, index=False, unique=False):
+                 primary_key=False, column_name=None, index=False,
+                 unique=False, default=None, extra_parameters=None):
         self.name = name
         self.field_class = field_class
         self.raw_column_type = raw_column_type
         self.nullable = nullable
         self.primary_key = primary_key
-        self.db_column = db_column
+        self.column_name = column_name
         self.index = index
         self.unique = unique
+        self.default = default
+        self.extra_parameters = extra_parameters
 
         # Foreign key metadata.
         self.rel_model = None
@@ -58,7 +64,7 @@ class Column(object):
             'raw_column_type',
             'nullable',
             'primary_key',
-            'db_column']
+            'column_name']
         keyword_args = ', '.join(
             '%s=%s' % (attr, getattr(self, attr))
             for attr in attrs)
@@ -66,22 +72,26 @@ class Column(object):
 
     def get_field_parameters(self):
         params = {}
+        if self.extra_parameters is not None:
+            params.update(self.extra_parameters)
 
         # Set up default attributes.
         if self.nullable:
             params['null'] = True
-        if self.field_class is ForeignKeyField or self.name != self.db_column:
-            params['db_column'] = "'%s'" % self.db_column
-        if self.primary_key and self.field_class is not PrimaryKeyField:
+        if self.field_class is ForeignKeyField or self.name != self.column_name:
+            params['column_name'] = "'%s'" % self.column_name
+        if self.primary_key and not issubclass(self.field_class, AutoField):
             params['primary_key'] = True
+        if self.default is not None:
+            params['constraints'] = '[SQL("DEFAULT %s")]' % self.default
 
         # Handle ForeignKeyField-specific attributes.
         if self.is_foreign_key():
-            params['rel_model'] = self.rel_model
+            params['model'] = self.rel_model
             if self.to_field:
-                params['to_field'] = "'%s'" % self.to_field
+                params['field'] = "'%s'" % self.to_field
             if self.related_name:
-                params['related_name'] = "'%s'" % self.related_name
+                params['backref'] = "'%s'" % self.related_name
 
         # Handle indexes on column.
         if not self.is_primary_key():
@@ -93,7 +103,7 @@ class Column(object):
         return params
 
     def is_primary_key(self):
-        return self.field_class is PrimaryKeyField or self.primary_key
+        return self.field_class is AutoField or self.primary_key
 
     def is_foreign_key(self):
         return self.field_class is ForeignKeyField
@@ -115,7 +125,12 @@ class Column(object):
 
     def get_field(self):
         # Generate the field definition for this column.
-        field_params = self.get_field_parameters()
+        field_params = {}
+        for key, value in self.get_field_parameters().items():
+            if isclass(value) and issubclass(value, Field):
+                value = value.__name__
+            field_params[key] = value
+
         param_str = ', '.join('%s=%s' % (k, v)
                               for k, v in sorted(field_params.items()))
         field = '%s = %s(%s)' % (
@@ -131,9 +146,11 @@ class Column(object):
 
 class Metadata(object):
     column_map = {}
+    extension_import = ''
 
     def __init__(self, database):
         self.database = database
+        self.requires_extension = False
 
     def execute(self, sql, *params):
         return self.database.execute_sql(sql, params)
@@ -144,30 +161,46 @@ class Metadata(object):
             for metadata in self.database.get_columns(table, schema))
 
         # Look up the actual column type for each column.
-        column_types = self.get_column_types(table, schema)
+        column_types, extra_params = self.get_column_types(table, schema)
 
         # Look up the primary keys.
         pk_names = self.get_primary_keys(table, schema)
         if len(pk_names) == 1:
             pk = pk_names[0]
             if column_types[pk] is IntegerField:
-                column_types[pk] = PrimaryKeyField
+                column_types[pk] = AutoField
+            elif column_types[pk] is BigIntegerField:
+                column_types[pk] = BigAutoField
 
         columns = OrderedDict()
         for name, column_data in metadata.items():
+            field_class = column_types[name]
+            default = self._clean_default(field_class, column_data.default)
+
             columns[name] = Column(
                 name,
-                field_class=column_types[name],
+                field_class=field_class,
                 raw_column_type=column_data.data_type,
                 nullable=column_data.null,
                 primary_key=column_data.primary_key,
-                db_column=name)
+                column_name=name,
+                default=default,
+                extra_parameters=extra_params.get(name))
 
         return columns
 
     def get_column_types(self, table, schema=None):
         raise NotImplementedError
 
+    def _clean_default(self, field_class, default):
+        if default is None or field_class in (AutoField, BigAutoField) or \
+           default.lower() == 'null':
+            return
+        if issubclass(field_class, _StringField) and \
+           isinstance(default, text_type) and not default.startswith("'"):
+            default = "'%s'" % default
+        return default or "''"
+
     def get_foreign_keys(self, table, schema=None):
         return self.database.get_foreign_keys(table, schema)
 
@@ -198,25 +231,52 @@ class PostgresqlMetadata(Metadata):
         1700: DecimalField,
         2950: TextField, # UUID
     }
+    array_types = {
+        1000: BooleanField,
+        1001: BlobField,
+        1005: SmallIntegerField,
+        1007: IntegerField,
+        1009: TextField,
+        1014: CharField,
+        1015: CharField,
+        1016: BigIntegerField,
+        1115: DateTimeField,
+        1182: DateField,
+        1183: TimeField,
+    }
+    extension_import = 'from playhouse.postgres_ext import *'
 
     def __init__(self, database):
         super(PostgresqlMetadata, self).__init__(database)
 
         if postgres_ext is not None:
             # Attempt to add types like HStore and JSON.
-            cursor = self.execute('select oid, typname from pg_type;')
+            cursor = self.execute('select oid, typname, format_type(oid, NULL)'
+                                  ' from pg_type;')
             results = cursor.fetchall()
 
-            for oid, typname in results:
-                if 'json' in typname:
+            for oid, typname, formatted_type in results:
+                if typname == 'json':
                     self.column_map[oid] = postgres_ext.JSONField
-                elif 'hstore' in typname:
+                elif typname == 'jsonb':
+                    self.column_map[oid] = postgres_ext.BinaryJSONField
+                elif typname == 'hstore':
                     self.column_map[oid] = postgres_ext.HStoreField
-                elif 'tsvector' in typname:
+                elif typname == 'tsvector':
                     self.column_map[oid] = postgres_ext.TSVectorField
 
+            for oid in self.array_types:
+                self.column_map[oid] = postgres_ext.ArrayField
+
     def get_column_types(self, table, schema):
         column_types = {}
+        extra_params = {}
+        extension_types = set((
+            postgres_ext.ArrayField,
+            postgres_ext.BinaryJSONField,
+            postgres_ext.JSONField,
+            postgres_ext.TSVectorField,
+            postgres_ext.HStoreField)) if postgres_ext is not None else set()
 
         # Look up the actual column type for each column.
         identifier = '"%s"."%s"' % (schema, table)
@@ -224,11 +284,15 @@ class PostgresqlMetadata(Metadata):
 
         # Store column metadata in dictionary keyed by column name.
         for column_description in cursor.description:
-            column_types[column_description.name] = self.column_map.get(
-                column_description.type_code,
-                UnknownField)
+            name = column_description.name
+            oid = column_description.type_code
+            column_types[name] = self.column_map.get(oid, UnknownField)
+            if column_types[name] in extension_types:
+                self.requires_extension = True
+            if oid in self.array_types:
+                extra_params[name] = {'field_class': self.array_types[oid]}
 
-        return column_types
+        return column_types, extra_params
 
     def get_columns(self, table, schema=None):
         schema = schema or 'public'
@@ -290,7 +354,7 @@ class MySQLMetadata(Metadata):
             name, type_code = column_description[:2]
             column_types[name] = self.column_map.get(type_code, UnknownField)
 
-        return column_types
+        return column_types, {}
 
 
 class SqliteMetadata(Metadata):
@@ -303,16 +367,19 @@ class SqliteMetadata(Metadata):
         'date': DateField,
         'datetime': DateTimeField,
         'decimal': DecimalField,
+        'float': FloatField,
         'integer': IntegerField,
         'integer unsigned': IntegerField,
         'int': IntegerField,
         'long': BigIntegerField,
+        'numeric': DecimalField,
         'real': FloatField,
         'smallinteger': IntegerField,
         'smallint': IntegerField,
         'smallint unsigned': IntegerField,
         'text': TextField,
         'time': TimeField,
+        'varchar': CharField,
     }
 
     begin = '(?:["\[\(]+)?'
@@ -342,7 +409,7 @@ class SqliteMetadata(Metadata):
         for column in columns:
             column_types[column.name] = self._map_col(column.data_type)
 
-        return column_types
+        return column_types, {}
 
 
 _DatabaseMetadata = namedtuple('_DatabaseMetadata', (
@@ -359,7 +426,8 @@ class DatabaseMetadata(_DatabaseMetadata):
         for index in self.indexes[table]:
             if len(index.columns) > 1:
                 field_names = [self.columns[table][column].name
-                               for column in index.columns]
+                               for column in index.columns
+                               if column in self.columns[table]]
                 accum.append((field_names, index.unique))
         return accum
 
@@ -372,7 +440,7 @@ class DatabaseMetadata(_DatabaseMetadata):
 
 
 class Introspector(object):
-    pk_classes = [PrimaryKeyField, IntegerField]
+    pk_classes = [AutoField, IntegerField]
 
     def __init__(self, metadata, schema=None):
         self.metadata = metadata
@@ -387,8 +455,10 @@ class Introspector(object):
             metadata = PostgresqlMetadata(database)
         elif isinstance(database, MySQLDatabase):
             metadata = MySQLMetadata(database)
-        else:
+        elif isinstance(database, SqliteDatabase):
             metadata = SqliteMetadata(database)
+        else:
+            raise ValueError('Introspection not supported for %r' % database)
         return cls(metadata, schema=schema)
 
     def get_database_class(self):
@@ -398,27 +468,40 @@ class Introspector(object):
         return self.metadata.database.database
 
     def get_database_kwargs(self):
-        return self.metadata.database.connect_kwargs
+        return self.metadata.database.connect_params
+
+    def get_additional_imports(self):
+        if self.metadata.requires_extension:
+            return '\n' + self.metadata.extension_import
+        return ''
 
     def make_model_name(self, table):
         model = re.sub('[^\w]+', '', table)
-        return ''.join(sub.title() for sub in model.split('_'))
+        model_name = ''.join(sub.title() for sub in model.split('_'))
+        if not model_name[0].isalpha():
+            model_name = 'T' + model_name
+        return model_name
 
     def make_column_name(self, column):
-        column = re.sub('_id$', '', column.lower()) or column.lower()
+        column = re.sub('_id$', '', column.lower().strip()) or column.lower()
+        column = re.sub('[^\w]+', '_', column)
         if column in RESERVED_WORDS:
             column += '_'
+        if len(column) and column[0].isdigit():
+            column = '_' + column
         return column
 
-    def introspect(self, table_names=None):
+    def introspect(self, table_names=None, literal_column_names=False,
+                   include_views=False):
         # Retrieve all the tables in the database.
-        if self.schema:
-            tables = self.metadata.database.get_tables(schema=self.schema)
-        else:
-            tables = self.metadata.database.get_tables()
+        tables = self.metadata.database.get_tables(schema=self.schema)
+        if include_views:
+            views = self.metadata.database.get_views(schema=self.schema)
+            tables.extend([view.name for view in views])
 
         if table_names is not None:
             tables = [table for table in tables if table in table_names]
+        table_set = set(tables)
 
         # Store a mapping of table name -> dictionary of columns.
         columns = {}
@@ -443,8 +526,16 @@ class Introspector(object):
                 foreign_keys[table] = self.metadata.get_foreign_keys(
                     table, self.schema)
             except ValueError as exc:
-                err(exc.message)
+                err(*exc.args)
                 foreign_keys[table] = []
+            else:
+                # If there is a possibility we could exclude a dependent table,
+                # ensure that we introspect it so FKs will work.
+                if table_names is not None:
+                    for foreign_key in foreign_keys[table]:
+                        if foreign_key.dest_table not in table_set:
+                            tables.append(foreign_key.dest_table)
+                            table_set.add(foreign_key.dest_table)
 
             model_names[table] = self.make_model_name(table)
 
@@ -452,7 +543,10 @@ class Introspector(object):
                                   for column_name in table_columns)
 
             for col_name, column in table_columns.items():
-                new_name = self.make_column_name(col_name)
+                if literal_column_names:
+                    new_name = re.sub('[^\w]+', '_', col_name)
+                else:
+                    new_name = self.make_column_name(col_name)
 
                 # If we have two columns, "parent" and "parent_id", ensure
                 # that when we don't introduce naming conflicts.
@@ -518,13 +612,17 @@ class Introspector(object):
             model_names,
             indexes)
 
-    def generate_models(self, skip_invalid=False, table_names=None):
-        database = self.introspect(table_names=table_names)
+    def generate_models(self, skip_invalid=False, table_names=None,
+                        literal_column_names=False, bare_fields=False,
+                        include_views=False):
+        database = self.introspect(table_names, literal_column_names,
+                                   include_views)
         models = {}
 
         class BaseModel(Model):
             class Meta:
                 database = self.metadata.database
+                schema = self.schema
 
         def _create_model(table, models):
             for foreign_key in database.foreign_keys[table]:
@@ -535,7 +633,7 @@ class Introspector(object):
 
             primary_keys = []
             columns = database.columns[table]
-            for db_column, column in columns.items():
+            for column_name, column in columns.items():
                 if column.primary_key:
                     primary_keys.append(column.name)
 
@@ -544,39 +642,54 @@ class Introspector(object):
 
             class Meta:
                 indexes = multi_column_indexes
+                table_name = table
 
             # Fix models with multi-column primary keys.
+            composite_key = False
             if len(primary_keys) == 0:
                 primary_keys = columns.keys()
             if len(primary_keys) > 1:
-                Meta.primary_key = CompositeKey([
+                Meta.primary_key = CompositeKey(*[
                     field.name for col, field in columns.items()
                     if col in primary_keys])
+                composite_key = True
 
             attrs = {'Meta': Meta}
-            for db_column, column in columns.items():
+            for column_name, column in columns.items():
                 FieldClass = column.field_class
-                if FieldClass is UnknownField:
+                if FieldClass is not ForeignKeyField and bare_fields:
+                    FieldClass = BareField
+                elif FieldClass is UnknownField:
                     FieldClass = BareField
 
                 params = {
-                    'db_column': db_column,
+                    'column_name': column_name,
                     'null': column.nullable}
-                if column.primary_key and FieldClass is not PrimaryKeyField:
+                if column.primary_key and composite_key:
+                    if FieldClass is AutoField:
+                        FieldClass = IntegerField
+                    params['primary_key'] = False
+                elif column.primary_key and FieldClass is not AutoField:
                     params['primary_key'] = True
                 if column.is_foreign_key():
                     if column.is_self_referential_fk():
-                        params['rel_model'] = 'self'
+                        params['model'] = 'self'
                     else:
                         dest_table = column.foreign_key.dest_table
-                        params['rel_model'] = models[dest_table]
+                        params['model'] = models[dest_table]
                     if column.to_field:
-                        params['to_field'] = column.to_field
+                        params['field'] = column.to_field
 
                     # Generate a unique related name.
-                    params['related_name'] = '%s_%s_rel' % (table, db_column)
-                if db_column in column_indexes and not column.is_primary_key():
-                    if column_indexes[db_column]:
+                    params['backref'] = '%s_%s_rel' % (table, column_name)
+
+                if column.default is not None:
+                    constraint = SQL('DEFAULT %s' % column.default)
+                    params['constraints'] = [constraint]
+
+                if column_name in column_indexes and not \
+                   column.is_primary_key():
+                    if column_indexes[column_name]:
                         params['unique'] = True
                     elif not column.is_foreign_key():
                         params['index'] = True
